//=================================================
/*	This is a program for playing notes, version 3.5.

	When specified file(s) as argument(s) in command
	line program plays this file(s), otherwise
	it works interactively.
*/
//-------------------------------------------------

#include <dos.h>
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <conio.h>

#define PAUSE 31416	// arbitrary big int >20000 and <32000

/* To find frequency corresponding to digit N you must say (simply case):
	freq[note_freq[num_note[N+19]]]
   In more complex case (using hashes) you should say:
	freq[note_freq[num_note[N+19]] +
	  num_hash[(N+19)%7]*hashflag[(N+19)%7] +
	  tact_hash[(N+19)%7]*tactflag[(N+19)%7]]
*/
#define MAXFREQ 101
#define MAXNOTE 56
float freq[MAXFREQ+1];	// note frequences; freq[0] is "do" of some octava
int note_freq[MAXNOTE+1];	// frequency number of note
int num_note[39];	// pointer to note
int num_hash[7];	// shift of note acting to the full line
int hashflag[7];	// flag to use (1) or not to use (0) num_hash
int tactflag[7];	// flag to use hash for current tact
int tact_hash[7];	// hash to tact
int sethash=0;	// to set tact_hash,tactflag & hashflag at next note
int set_tact_hash;	//\
int set_tactflag;	//  values setting on next note
int set_hashflag;	///

int freqdist[7] = {2,2,1,2,2,2,1};	// distance between notes in polutones
int violine_shift = 30;	// "0" in skripichny clef ("mi")
int bass_shift = 25;	// "0" in bass clef ("sol")

float long_factor = 1.;	// factor to multiply longitude
float accel = 1.;	// longitude acceleration

int bracket;	// bracket has been open (for micropause to be absent)
int bracketnext;	// to set bracket at next character

int begin_of_file = 1;	// before reading a file

char c;	// current character

FILE *fp;	// current file

int dig = 0;	// previous char was a digit
int hight;	// sound hight. Lower line is 0, next up is 2 etc.
const int longdef = 4;	// default longitude

char line[256];	// line of text
int line_pointer;	// current position in the line

int read_macro = 0;	// at the beginning of the macro

int new_line = 1;	// now is the beginning of the line
int new_line_sav = 0;	// now is the beginning of the line

#define DO_DIVIDER 16	// divider to go to the lowest do, is 2**4
float freq0 = 150./DO_DIVIDER;	// starting frequency, Hz, lowest "do"

int long0 = 125;	// msec of 1/16
float accel_factor = 1.3;	// acceleration of the note
int accel_count = 3;	// count of accelerations
int accel_remain = 0;	// count of remain accelerations
int micropause = 30;	// pause between same notes


//--- functions used:

void playfile(void);
void init_freq(void);
int clef(void);
int lineset(void);
int isunidigit(int *digit);
//int isunidigit(void);

int isunibracket(void);
int isunilong(int *longitude);
int isunihash(void);
int isuniaccel(void);
int playline(void);
void note(int hight, int longitude);
int getsym(void);
int readline(void);
void num_note_set(int shift);
void num_note_shift(int shift);
void defaults(void);
void dig_note(void);
int isrdigit(char cd);
int isralpha(char cd);


#include "notmac3.cpp"

//==================================================
	int main(int argc, char *argv[])
//--------------------------------------------------
{

//.................................................

init_freq();
init_saveline();

if(argc == 1) {	// input from stdin
  fp = stdin;
  finp[0] = fp;	// for stack of input files
  playfile();
} else {
  while(--argc) {
    begin_of_file = 1;
    if((fp = fopen(*++argv,"r")) == NULL) {
      printf("Sorry, I can't open file %s.\n", *argv);
      continue;
      }
    finp[0] = fp;	// for stack of input files
    playfile();
    fclose(fp);
    delay(2000);	// pause between files
  }
}

return 0;

}

//=================================================
	void playfile(void)
//-------------------------------------------------
//	Plays contents of the file IFP.
//-------------------------------------------------
{

defaults();

//--- getting started
//printf("before getsym in playfile\n");
if(getsym()==-1) return;
//printf("after getsym in playfile\n");
if(playline()==-1) return;

}


//=================================================
	void init_freq(void)
//-------------------------------------------------
//	Initializes note frequences.
//	0 is "mi", not a "do".
//-------------------------------------------------
{

//float freq0 = 150./16.;	// Hz, lowest "do"
//int freq0 = 183.3;	// Hz, "mi"
float epsilon = 0.00;	// correction

float polutone;

polutone = pow(2.,1./12.)+epsilon;

freq[0] = freq0;

for(int i=0;i<MAXFREQ-1;i++) {
  freq[i+1] = freq[i]*polutone;
  }

note_freq[0] = 0;

for(i=0;i<8;i++) {
  for(int j=0;j<7;j++) {
    note_freq[i*7+j+1] = note_freq[i*7+j] + freqdist[j];
    }
  }

shifted = 0;	// for macro \shift{}

}

//=================================================
	int clef(void)
//-------------------------------------------------
//	If "c" is clef this routine sets frequences
//	and returns 1, otherwise returns 0.
//
//-------------------------------------------------
{
int isclef;

int shift;

//.................................................

isclef = 0;

if(c=='&') {	// violine clef
  shift = violine_shift;
  isclef = 1;
  }

if(c=='~') {	// bass clef
  shift = bass_shift;
  isclef = 2;
  }

if(isclef) {
  num_note_set(shift);
  if(getsym()==-1) return -1;
  return isclef;
} else {
  return 0;
  }

}

//=================================================
	void num_note_set(int shift)
//-------------------------------------------------
/*	Sets hashes
*/
//-------------------------------------------------
{
for(int i=0;i<39;i++) {
  num_note[i] = i-19 + shift;
  }
for(i=0;i<7;i++) {
  num_hash[i] = 0;
  tact_hash[i] = 0;
  }

}

//=================================================
	void num_note_shift(int shift)
//-------------------------------------------------
/*
	Shifts notes on "shift" expressed in notes.
	shift = 7 is octava.
*/
//-------------------------------------------------
{
for(int i=0;i<39;i++) {
  num_note[i] += shift;
  }

}

//=================================================
	int lineset(void)
//-------------------------------------------------
/*	Sets hashes
*/
//-------------------------------------------------
{
int hash;
int digit;
//.................................................

hash = 0;

loop:

//printf("You are in lineset\n");

if(c=='#') {	// sharp
  if(getsym()==-1) return -1;
//  if(hash==1) return 1;
  hash = 1;
} else if(c=='%') {	// flat
  if(getsym()==-1) return -1;
//  if(hash==-1) return 1;
  hash = -1;
  }

//if(isunidigit(*digit)==-1) return -1;
switch(isunidigit(&digit)) {
//switch(isunidigit()) {
  case -1:	return -1;
  case 0:	break;
  default:	if(digit==PAUSE) {	// change this line
		  if(hash) return 2;
		  defaults();
		  }
		//printf("digit in lineset = %d, hash = %d\n",digit,hash);
		num_hash[(digit+19)%7] = hash;
		hashflag[(digit+19)%7] = 1;
		goto loop;
  }

if(isspace(c)) {
  return 1;
} else {
  printf("\nYou should end clef settings by space.\n");
  }

return 0;

}

//=================================================
	void defaults(void)
//-------------------------------------------------
/*
	Sets defaults.

*/
//-------------------------------------------------
{

num_note_set(violine_shift);

for(int i=0;i<7;i++) {
  hashflag[i] = 1;
  tactflag[i] = 0;
  }

sethash=0;

long_factor = 1.;
accel = 1.;

bracket = 0;
bracketnext = 0;

dig = 0;

soundonoff = 1;

cpause = '?';	// character for pause

stack_input = 1;
stack_input_sav = 1;
immediate = 1;
immediate_sav = 1;

freq0 = 150./DO_DIVIDER;	// starting frequency, Hz, lowest "do"
init_freq();
long0 = 125;
accel_factor = 1.3;
accel_count = 3;
accel_remain = 0;

strcpy(editor,"");
strcpy(norton,"");

input_ini();

}


//=================================================
	int isunidigit(int *digit)
//	int isunidigit(void)
//-------------------------------------------------
/*	Says whether current character is unified
	digit and returns its value as "digit".

	Unified digit may have following forms:
		     3 =   3
		    -3 =  -3
		    /3 =  13
		   -/3 = -13
	    and also ? = PAUSE (you may reset this by macro \pause{x},
				where x is new pause character)

	Returns: 0 if "c" was not a digit
		 1 if "c" was a digit
		-1 if "c" was a digit and EOF read
		-2 if "c" was not a full unified digit
		   and EOF read.
*/
//-------------------------------------------------
{
int sign;

//.................................................

sign = 1;

if(c==cpause) {
  *digit = PAUSE;
  if(getsym()==-1) return -1;
  return 1;
  }

after_minus:

if(isrdigit(c)) {
  *digit = (c-48)*sign;
//  digit = (c-48)*sign;
  if(getsym()==-1) return -1;
  return 1;
  }

if(c=='/') {
  if(getsym()==-1) return -2;
  if(isrdigit(c)) {
    *digit = (10 + (c - 48)) * sign;
//    digit = (10 + (c - 48)) * sign;
    if(getsym()==-1) return -1;
    return 1;
    }
  printf("\nWarning:unknown digit: %c\n.",c);
  return 0;
  }

if(c=='-') {
  if(sign!=1) printf("\nWarning: too many minus signs.\n");
  sign = -1;
  if(getsym()==-1) return -2;
  goto after_minus;
  }

return 0;

}

//==================================================
	int isunibracket(void)
//--------------------------------------------------
/*	Returns 1 when current character was a bracket
	and sets variable bracket to 1 (open) or
	0 (close), otherwise returns 0.
*/
//--------------------------------------------------
{

if(c=='(') {
  bracketnext = 1;	// to set bracket at next character
  if(getsym()==-1) return -1;
  return 1;
  }

if(c==')') {
  bracket = 0;
  if(getsym()==-1) return -1;
  return 1;
  }

return 0;

}


//==================================================
	int isunilong(int *plong)
//--------------------------------------------------
/*	Returns 1 when current character(s) was
	unified longitude, otherwise 0.

	Unified longitudes (returned by pointer "plong") may be:

		` = 1/16

		' = 1/8

		^ = 3/16

		, = 1/4

		" = 3/8

		. = 1/2

		; = 3/4

		: = 1

		! = 3/2

		+ = " = 3/4 when single

		+ after longitude multiplies it on 3/2

*/
//--------------------------------------------------
{

switch(c) {

  case '`': *plong = 1; break;

  case '\'': *plong = 2; break;

  case '^': *plong = 3; break;

  case ',': *plong = 4; break;

  case '"': *plong = 6; break;

  case '.': *plong = 8; break;

  case ';': *plong = 12; break;

  case ':': *plong = 16; break;

  case '!': *plong = 24; break;

  case '+': *plong = 6; if(getsym()==-1) return -1; return 1;

  default: return 0;

  }

if(getsym()==-1) return -1;

if(c=='+') {
  *plong *= 3./2.;
  if(getsym()==-1) return -1;
  }

return 1;

}


//==================================================
	int isunihash(void)
//--------------------------------------------------
/*	Returns 1 when current character was a hash
	or percent or dollar or "|" and sets variables
	"hashflag" and "tact_hash", otherwise returns 0.
*/
//--------------------------------------------------
{

switch(c) {
  case '#':	dig_note();
		sethash = 1;
		set_tact_hash = 1;
		set_tactflag = 1;
		set_hashflag = 0;
		break;
  case '%':	dig_note();
		sethash = 1;
		set_tact_hash = -1;
		set_tactflag = 1;
		set_hashflag = 0;
		break;
  case '$':	dig_note();
		sethash = 1;
		set_tact_hash = 0;
		set_tactflag = 0;
		set_hashflag = 0;
		break;
  case '|':	dig_note();
		sethash = 0;
		for(int i=0;i<7;i++) {
		  tact_hash[i] = 0;
		  tactflag[i] = 0;
		  hashflag[i] = 1;
		  }
		break;
  default:	return 0;
  }

if(getsym()==-1) return -1;

return 1;

}


//==================================================
	int isuniaccel(void)
//--------------------------------------------------
/*	Returns 1 when current character was a
	"<" or ">" or "=" or "*" and sets variable
	"accel" or resets "long_factor",
	otherwise returns 0.

	"<" is used for acceleration of music,
	">" for deceleration,
	"=" for fixing accelerated or decelerated
		unit of longitude and
	"*" for reseting longitude unit.
*/
//--------------------------------------------------
{

switch(c) {
  case '<': 	dig_note();
		accel = accel_factor;
		accel_remain = accel_count;
		break;	// faster
  case '>': 	dig_note();
		accel_remain = accel_count;
		accel = 1/accel_factor;
		break;	// slower
  case '=': 	dig_note();
		accel = 1.;
		break;		// constant
  case '*':	dig_note();			// reset
		accel = 1.;
		long_factor = 1.;
		break;
  default: return 0;
  }

if(getsym()==-1) return -1;

switch(isuniaccel()) {
  case -1:	return -1;
  case 0:       break;
  default:	long_factor /= accel;
  }

return 1;

}


//==================================================
	void dig_note(void)
//--------------------------------------------------
/*
	Plays last not if it was not yet played

*/
//--------------------------------------------------
{

if(dig) {
//  printf("dig_note works\n");
  note(hight,longdef);
  dig = 0;
  printf(line);
  line[0] = 0;
  line_pointer = 0;
  }
}

//=================================================
	int playline(void)
//-------------------------------------------------
//	Plays contents of the file IFP.
//
//Example: 1232.3,4'5;32:5::2 3:.2,'2`
//-------------------------------------------------
{
//int dig = 0;	// previous char was a digit

//int hight;	// sound hight. Lower line is 0, next up is 2 etc.
int digit;	// next hight, if any
int longitude;	// sound longitude in 1/16's
//const int longdef = 4;	// default longitude

//int first;

//.................................................

next:

//printf("next in playline\n");

switch(isuniaccel()) {
  case -1:	return -1;
  case 0:       break;
  default:	goto next;
  }

switch(isunihash()) {
  case -1:	return -1;
  case 0:       break;
  default:	goto next;
  }

//if(isunidigit(*digit)) {
switch(isunidigit(&digit)) {
//switch(isunidigit()) {
  case -1:	//if(dig) note(hight,longdef);
		dig_note();
		//printf("case -1 in playline, dig = %d\n",dig);
		note(digit,longdef);
  case -2:	return -1;
  case 0:	break;
  default:      //printf("case 1 in playline, dig = %d\n",dig);
		dig_note();
		dig = 1;
		hight = digit;
		goto next;
}

if(c=='\n') {
  dig_note();
  if(getsym()==-1) return -1;
  goto next;
  }

//if (isunilong(*longitude)) {
switch(isunilong(&longitude)) {
  case -1:	dig = 0;
		printf(line);
		line[0] = 0;
		line_pointer = 0;
		note(hight,longitude);
		return -1;
  case 0:	break;
  default:	dig = 0;
		printf(line);
		line[0] = 0;
		line_pointer = 0;
		note(hight,longitude);
		goto next;
  }

//if(isunibracket()) goto next;
switch(isunibracket()) {
  case -1:	return -1;
		break;
  case 0:	break;
  default:	goto next;
  }


printf("\nSome error: unknown character %c, code %d\n",c,c);
if(getsym()==-1) return -1;

goto next;

}


//=================================================
	void note(int hight, int longitude)
//-------------------------------------------------
//	Plays single sound.
//-------------------------------------------------
{


float frequency;	// frequency played
int longcorr;	// playing longitude corrected to micropause

int nfreq,nnote;
int h19 = hight+19;
//.................................................

//printf("hight:%d, long:%d\n",hight,longitude);
//printf("freq[hight+9]: %f\n",freq[hight+9]);

if(set_tact_hash) {
  tact_hash[h19%7] = set_tact_hash;
  set_tact_hash = 0;
  }

if(sethash) {
  hashflag[h19%7] = set_hashflag;
  tactflag[h19%7] = set_tactflag;
  sethash = 0;
  }

/* To find frequency corresponding to digit N you must say (simply case):
	freq[note_freq[num_note[N+19]]]
   In more complex case (using hashes) you should say:
	freq[note_freq[num_note[N+19]] +
	  num_hash[(N+19)%7]*hashflag[(N+19)%7] +
	  tact_hash[(N+19)%7]*tactflag[(N+19)%7]]
*/

//--- define longitude

if(accel_remain<=0) {
  accel = 1;
} else {
  accel_remain--;
  }

long_factor /= accel;

longcorr = longitude*long0*long_factor;

if(!bracket) {
  longcorr -= micropause;
  if(longcorr<0) {
    printf("\nError: note longitude less than 0. Check micropause.\n");
    longcorr = 100;
    }
  }

//--- playing

if(hight!=PAUSE) {
  nnote = num_note[h19];
  if(nnote<0||nnote>MAXNOTE) {
    printf("\nError: frequency is outside the available range.\n");
    return;
    }
  nfreq = note_freq[nnote] + num_hash[h19%7]*hashflag[h19%7] + tact_hash[h19%7]*tactflag[h19%7];
  if(nfreq<0||nfreq>MAXFREQ) {
    printf("\nError: frequency is outside the available range.\n");
    return;
    }
  frequency = freq[nfreq];
  if(!soundonoff) {
    printf("frequency:%f, longitude:%d\n",frequency,longcorr);
  } else {
    sound(frequency);
    }
} else {
  nosound();
  }

delay(longcorr);

//nosound();

//printf("after delay in note, bracket = %d\n",bracket);

if(!bracket) {
  nosound();
  delay(micropause);
  }
//printf("after nosound in note, bracket = %d\n",bracket);

//--- set bracket

if(bracketnext) {
  bracketnext = 0;
  bracket = 1;
  }

}


//==================================================
	int isrdigit(char cd)
//==================================================
/*	Determines whether cd is digit. Not fails
	if cd is russian letter.
*/
//--------------------------------------------------
{
return (cd>=48&&cd<58);

}


//==================================================
	int isralpha(char cd)
//==================================================
/*	Determines whether cd is a alphabetic letter
	including russian letters.
*/
//--------------------------------------------------
{
int letter = 0;
//..................................................

if(cd>=-128&&cd<=-81) letter = 1;	// - + -

else if(cd>=-32&&cd<=-17) letter = 1;	// -

else if(isalpha(cd)) letter = 1;

else if(cd==' ') letter = 1;

else if(cd=='_') letter = 1;

if(cd==cpause) letter = 0;

return letter;

}


//==================================================
	int getsym(void)
//--------------------------------------------------
/*	Performs the following operations:

	- reads next character ("c") from the file line

	- when "c" is alphabetical outputs it to stdout
	  and reads new character

	- when "c" is "\" reads following character
	  and if it is not alphabetical sends it
	  literally to "stdout", otherwise calls
	  macro conserning to the word following the "\"
	  (last yet does not work, but only one character
	  outputs); after this reads new character

	- when EOL found:

		- calls playing of the note if last
		  character was a digit

		- reads next line

		- at the beginning of the line performs
		  line settings, if any

		- skipes line when first character
		  is "!"

		- calls playing another file when line
		  begins from "@"

	- at the beginning of the file reads
	  first line

	Returns: character read or -1 if EOF found

*/
//--------------------------------------------------
{
//char line[256];	// line of input file
char lineformat[10],linelen[5];

char command0[100], command[256];	// call to this program

int skeep_line = 0;
int read_EOL;
int command_pointer;

int backslash = 0;	// flag for backslash

int tmp;

//.................................................

strcpy(command0,_argv[0]);
strcat(command0," ");

/*
itoa(sizeof(line),linelen,10);
strcat(lineformat,"%");
strcat(lineformat,linelen);
strcat(lineformat,"s");
*/

backslash = 0;
read_EOL = 0;

//printf("1: new_line = %d, backslash = %d\n",new_line,backslash);

if(begin_of_file) {
  //printf("At the beginning of the file...\n");
  num_note_set(violine_shift);
  begin_of_file = 0;
  new_line = 1;
  }

//printf("2: new_line = %d, backslash = %d\n",new_line,backslash);

//if(read_macro) {
//  callmacro();
//  read_macro = 0;
//  }

next_char:

//c = getc(fp);
c = getsymbol();

next_char1:

if(c==EOF) {
  if(read_EOL) {
    command[command_pointer++] = 0;
    system(command);
    }
  return -1;
  }

//printf("Read %c, code %d\n",c,c);

//printf("3: new_line = %d, backslash = %d\n",new_line,backslash);

if(c=='\n') {
  //putchar(c);
  if(dig) {
    line[line_pointer++] = c;
    line[line_pointer] = 0;
  } else {
    if(!skeep_line) putchar(c);
    }
  new_line = 1;
  //printf("At the beginning of the line...\n");
  skeep_line = 0;
  if(read_EOL) {
    command[command_pointer++] = 0;
    system(command);
    }
  read_EOL = 0;
//  goto next_char;
  return 0;
} else {
  if(skeep_line) goto next_char;
  if(read_EOL) {
    command[command_pointer++] = c;
    goto next_char;
    }
  }

//printf("4: new_line = %d, backslash = %d\n",new_line,backslash);

if(backslash) {
  //printf("C is after backslash.\n");
//  if(isalpha(c)||c=='_') {
  if(isralpha(c)||c=='_') {
    callmacro();
    //read_macro = 1;
    backslash = 0;
    goto next_char1;
  } else {
    if(dig) {
      line[line_pointer++] = c;
      line[line_pointer] = 0;
    } else {
      putchar(c);
      }
    }
  //if(c) putchar(c);
  //else printf("\nYou should not type backslash (\) at the end of the line\n");
  backslash = 0;
  goto next_char;
  }

//printf("5: new_line = %d, backslash = %d\n",new_line,backslash);

//--- check on new line

if(new_line) {
  //printf("At the begining of the line (2)...\n");
  new_line = 0;
//  printf("6: new_line = %d, backslash = %d\n",new_line,backslash);
  skeep_line = 0;
  if(c=='!') {	// ignore current line
    //printf("Comment line...\n");
    //    fscanf(fp,lineformat,line);
    skeep_line = 1;
    goto next_char;
    }
  if(c=='@') {	// play new file
    //printf("@ line...\n");
    strcpy(command,command0);
    command_pointer = strlen(command);
    read_EOL = 1;
    goto next_char;
    }

//  if(clef()) lineset();
  tmp = clef();
  //printf("clef is %d\n",tmp);
  switch (tmp) {
    case 0:    	break;
    case -1:	return -1; break;
    default:	if(lineset()==-1) return -1;
    }
  //printf("After clef switch\n");
  }

//if(isalpha(c)||c=='_') {
if(isralpha(c)||c=='_') {
//if(isascii(c)) {
  //printf("C is alphabetic...\n");
  if(dig) {
    line[line_pointer++] = c;
    line[line_pointer] = 0;
  } else {
    putchar(c);
    }
  goto next_char;
  }

if(c=='\\') {
  //printf("C is backslash...\n");
  backslash = 1;
  goto next_char;
  }

//printf("At the end of the getsym...\n");

return 0;

}
